#!/usr/bin/env python
# Copyright 2014-2015 ARM Limited
#
# Licensed under the Apache License, Version 2.0
# See LICENSE file for details.

# standard library modules, , ,
import sys
import subprocess
import os

def run(arguments, cwd='.'):
    progdir = os.path.join(os.path.dirname(os.path.realpath(__file__)), '..', '..')
    runprog = [
        sys.executable,
        '-c',
        "import sys; sys.path.insert(0, '%s'); import valinor; valinor.main()" % progdir
    ]
    child = subprocess.Popen(
          args = runprog + arguments,
           cwd = cwd,
        stdout = subprocess.PIPE,
        stderr = subprocess.PIPE,
         stdin = subprocess.PIPE
    )
    out, err = child.communicate()
    return out.decode('utf-8'), err.decode('utf-8'), child.returncode







# Copyright 2015 ARM Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import argparse
import os
import sys
import pkg_resources
import shutil

import valinor.logging_setup as logging_setup
import valinor.ide_detection as ide_detection
import valinor.elf as elf
from project_generator.project import Project
from project_generator.generate import Generator
from project_generator.settings import ProjectSettings

def main():
    logging_setup.init()
    logging.getLogger().setLevel(logging.INFO)

    p = argparse.ArgumentParser()

    p.add_argument('--version', dest='show_version', action='version',
        version=pkg_resources.require("valinor")[0].version,
        help='display the version'
    )

    p.add_argument('-t', '--tool', dest='ide_tool', default=None,
        help='Debug tool (IDE) to generate for. If omitted, a debug project '+
             'will be generated for an IDE detected on your system, '+
             'defaulting to opening a GDB debug session, if no known IDEs '+
             'are detected'
    )

    p.add_argument('-d', '--project-dir', dest='project_dir', default=None,
        help='The directory in which to generate any necessary project files.  '+
             'Defaults to the directory of the executable argument.'
    )

    p.add_argument('-n', '--no-open', dest='start_session', default=True, action='store_false',
        help='Do not open the debug session, just generate the necessary '+
             'files to enable debugging, and print the command that would be '+
             'necessary to proceed.'
    )

    p.add_argument('--target', dest='target', required=True,
        help='The target board to generate a project file for (e.g. K64F).'
    )

    p.add_argument('executable',
        help='Path to the executable to debug.'
    )

    args = p.parse_args()

    # check that the executable exists before we proceed, so we get a nice
    # error message if it doesn't
    if not os.path.isfile(args.executable):
        logging.error('cannot debug file "%s" that does not exist' % args.executable)
        sys.exit(1)

    # Get setttings and generator (it updates targets def prior select)
    projects = {
        'projects' : {}
    }
    generator = Generator(projects)
    project_settings = ProjectSettings()

    available_ides = ide_detection.available()
    ide_tool = args.ide_tool
    if not ide_tool:
        ide_tool = ide_detection.select(available_ides, args.target, project_settings)
        if ide_tool is None:
            if len(available_ides):
                logging.error('None of the detected IDEs supports "%s"', args.target)
            else:
                logging.error('No IDEs were detected on this system!')
            logging.info('Searched for:\n  %s', '\n  '.join(ide_detection.IDE_Preference))
    if ide_tool is None:
        logging.error(
            'No IDE tool available for target "%s". Please see '+
            'https://github.com/project-generator/project_generator for details '+
            'on adding support.', args.target
        )
        sys.exit(1)

    file_name      = os.path.split(args.executable)[1]
    file_base_name = os.path.splitext(file_name)[0]
    executable_dir = os.path.dirname(args.executable)

    projectfile_dir = args.project_dir or executable_dir

    files = elf.get_files_from_executable(args.executable)

    # pass empty data to the tool for things we don't care about when just
    # debugging (in the future we could add source files by reading the debug
    # info from the file being debugged)
    project_data = {
        'common': {
            'target': [args.target],  # target
            'build_dir': ['.'],
            'linker_file': ['None'],
            'export_dir': ['.' + os.path.sep + projectfile_dir],
            'output_dir': {
                'rel_path' : [''],
                'path' : [os.path.relpath(executable_dir, projectfile_dir) + os.path.sep],
            },
            'sources': {'Source_Files':sorted([f for f in files], key=lambda file: os.path.basename(file))},
        }
    }

    project = Project(file_base_name, [project_data], project_settings)
    project.generate(ide_tool)

    # perform any modifications to the executable itself that are necessary to
    # debug it (for example, to debug an ELF with Keil uVision, it must be
    # renamed to have the .axf extension)
    executable = args.executable
    if ide_tool in ('uvision', 'uvision5'):
        new_exe_path = args.executable + '.axf'
        shutil.copy(args.executable, new_exe_path)
        executable = new_exe_path
    projectfiles = project.get_generated_project_files(ide_tool)
    if not projectfiles:
        logging.error("failed to generate project files")
        sys.exit(1)

    if args.start_session:
        launch_fn = ide_detection.get_launcher(ide_tool)
        if launch_fn is not None:
            try:
                launch_fn(projectfiles['files'], executable)
            except Exception as e:
                logging.error('failed to launch debugger: %s', e)
        else:
            logging.warning('failed to open IDE')
    print('project files have been generated in: %s' % os.path.join(os.getcwd(), os.path.normpath(projectfiles['path'])))



#!/usr/bin/env python
# Copyright 2015 ARM Limited
#
# Licensed under the Apache License, Version 2.0
# See LICENSE file for details.

import subprocess
import logging
import threading
import signal
import traceback
try:
    import Queue
except ImportError:
    import queue as Queue

logger = logging.getLogger('gdb')

def _ignoreSignal(signum, frame):
    logging.debug('ignoring signal %s, traceback:\n%s' % (
        signum, ''.join(traceback.format_list(traceback.extract_stack(frame)))
    ))

def _launchPyOCDGDBServer(msg_queue):
    logger.info('preparing PyOCD gdbserver...')
    from pyOCD.gdbserver import GDBServer
    from pyOCD.board import MbedBoard
    gdb = None
    try:
        logger.info('finding connected board...')
        board_selected = MbedBoard.chooseBoard(blocking=False)
        if board_selected is not None:
            with board_selected as board:
                logger.info('starting PyOCD gdbserver...')
                gdb = GDBServer(
                    board, 3333, {
                         'break_at_hardfault': True,
                        'step_into_interrupt': False,
                             'break_on_reset': False,
                    }
                )
                if gdb.isAlive():
                    msg_queue.put('alive')
                    while gdb.isAlive():
                        gdb.join(timeout = 0.5)
                        # check for a "kill" message from the parent process:
                        try:
                            msg = msg_queue.get(False)
                            if msg == 'kill':
                                gdb.stop()
                                break
                        except Queue.Empty:
                            pass
        else:
            logger.error('failed to find a connected board')
    except Exception as e:
        logger.error('exception in GDB server thread: %s', e)
        if gdb != None:
            gdb.stop()
        raise
    msg_queue.put('dead')

def launcher(gdb_exe):
    def launch_gdb(projectfiles, executable):
        # the "projectfiles" for gdb are really command files that we should
        # execute to set up the debug session:
        cmd = [gdb_exe]
        for f in projectfiles:
            cmd += ['-x', f]
        cmd.append(executable)
        # ignore Ctrl-C while gdb is running:
        signal.signal(signal.SIGINT, _ignoreSignal);
        child = subprocess.Popen(cmd)
        child.wait()
    return launch_gdb

def arm_none_eabi_launcher(gdb_exe):
    gdb_launcher = launcher(gdb_exe)
    def launch_arm_gdb(projectfiles, executable):
        queue = Queue.Queue()
        t = threading.Thread(target=_launchPyOCDGDBServer, args=(queue,))
        try:
            t.start()
            # wait for an 'alive' message from the server before starting gdb
            # itself:
            msg = None
            while msg != 'alive':
                try:
                    msg = queue.get(timeout=1.0)
                except Queue.Empty as e:
                    msg = None
                    pass
                if msg == 'dead' or not t.is_alive():
                    raise Exception('gdb server failed to start')
        except KeyboardInterrupt as e:
            logger.error('stopped by user')
            queue.put('kill')
            t.join()
            raise
        gdb_launcher(projectfiles, executable)
        queue.put('kill')
        t.join()
    return launch_arm_gdb





#!/usr/bin/env python
# Copyright 2015 ARM Limited
#
# Licensed under the Apache License, Version 2.0
# See LICENSE file for details.

import subprocess
import logging
import os

from distutils.spawn import find_executable

from project_generator import tools_supported
from project_generator_definitions.definitions import ProGenDef

from valinor.gdb import launcher as gdb_launcher
from valinor.gdb import arm_none_eabi_launcher as arm_none_eabi_gdb_launcher

# cache the detected IDEs, (map of ide name to a function(projectfiles,
# executable) that will launch it)
IDE_Launchers = {
}

IDEs_Scanned = False

# preferred order of IDEs if multiple are available
IDE_Preference = [
    'uvision', 'uvision5', 'arm_none_eabi_gdb', 'gdb'
]


logger = logging.getLogger('ide_detect')

def _read_hklm_reg_value(key_path, value_name):
    ''' read a value from a registry key under HKEY_LOCAL_MACHINE '''
    if os.name != 'nt':
        return None
    import _winreg as winreg

    k = None
    value = None
    try:
        k = winreg.OpenKey(winreg.HKEY_LOCAL_MACHINE, key_path)
        v = winreg.QueryValueEx(k, value_name)
        return v[0]
    except WindowsError:
        if k:
            winreg.CloseKey(k)
    return value

def _find_uvision():
    found = find_executable('UV4')
    if found: return found
    if os.name == 'nt':
        found_pathdir = _read_hklm_reg_value(r'Software\Keil\Products\MDK', 'Path')
        if not found_pathdir:
            found_pathdir = _read_hklm_reg_value(r'Software\Wow6432Node\Keil\Products\MDK', 'Path')
        if found_pathdir:
            found = os.path.join(found_pathdir, '..', 'UV4', 'UV4.exe')
            if os.path.isfile(found):
                return os.path.normpath(found)
    return None


def _find_generic_gdb():
    return find_executable('gdb')


def _find_arm_none_eabi_gdb():
    return find_executable('arm-none-eabi-gdb')


def _uvision_launcher(uvision_exe):
    def launch_uvision(projectfiles, executable):
        uvproj = [
            x for x in projectfiles if x.endswith('.uvproj') or x.endswith('.uvprojx')
        ]
        if len(uvproj) != 1:
            raise Exception('Exactly one project file must be provided')
        logger.info("launching uvision: %s %s", uvision_exe, uvproj[0])
        child = subprocess.Popen(
            [uvision_exe, uvproj[0]],
        )
        child.wait()
    return launch_uvision

IDE_Scanners = {
              'uvision': (_find_uvision, _uvision_launcher),
             'uvision5': (_find_uvision, _uvision_launcher),
                  'gdb': (_find_generic_gdb, gdb_launcher),
    'arm_none_eabi_gdb': (_find_arm_none_eabi_gdb, arm_none_eabi_gdb_launcher),
}

def _ensure_IDEs_scanned():
    global IDEs_Scanned, IDE_Launchers
    if IDEs_Scanned:
        return

    # !!! could scan in parallel, should be worth it as scanners are
    # likely highly io-bound
    for ide, (scanner, launcher) in IDE_Scanners.items():
        logger.debug('scanning for %s: %s', ide, (scanner, launcher))
        detected = scanner()
        logger.debug('scanning for %s... %sfound', ide, ('not ', '')[bool(detected)])
        if detected:
            IDE_Launchers[ide] = launcher(detected)
            # make sure we have a preference order for all detected IDEs, even
            # if the list hasn't been updated:
            if ide not in IDE_Preference:
                IDE_Preference.append(ide)

    IDEs_Scanned = True


def available():
    ''' return a list of available IDEs on this platform '''
    _ensure_IDEs_scanned()
    return [k for k in IDE_Launchers.keys()]


def select(available_ides, target, project_settings):
    ''' select the preferred option out of the available IDEs to debug the
    selected target, or None '''
    possible_ides = []
    for ide in available_ides:
        tool = tools_supported.ToolsSupported().get_tool(ide)
        if not tool.is_supported_by_default(target):
            if ProGenDef(ide).is_supported(target):
                possible_ides.append(ide)
        else:
            possible_ides.append(ide)

    if len(possible_ides):
        return sorted(possible_ides, key=lambda x:IDE_Preference.index(x))[0]
    else:
        return None


def get_launcher(ide):
    ''' return a function(projectfiles) that will launch the
    specified IDE when called, or None if that IDE cannot be found '''
    if ide in IDE_Launchers:
        return IDE_Launchers[ide]
    else:
        return None



# Copyright 2015 ARM Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import logging

from elftools.elf.elffile import ELFFile
from elftools.common.exceptions import ELFError

def get_files_from_executable(filename):
    with open(filename, 'rb') as f:
        # ELFFile looks for magic number, if there's none, ELFError is raised
        try:
            elffile = ELFFile(f)
        except ELFError:
            logging.info("%s is invalid elf file" % filename)
            return []

        if not elffile.has_dwarf_info():
            logging.info("File does not have dwarf info, no sources in the project file")
            return
        dwarfinfo = elffile.get_dwarf_info()

    files = []
    # Go over all the line programs in the DWARF information and get source files paths
    for CU in dwarfinfo.iter_CUs():
        top_DIE = CU.get_top_DIE()
        files.append(top_DIE.get_full_path())
    return files



# Copyright 2014-2015 ARM Limited
#
# Licensed under the Apache License, Version 2.0
# See LICENSE file for details.

import os
from setuptools import setup, find_packages

def read(fname):
    with open(os.path.join(os.path.dirname(__file__), fname)) as f:
        return f.read()
    return None

setup(
    name = "valinor",
    version = "1.1.4",
    author = 'Martin Kojtal, James Crosby',
    author_email = "c0170@rocketmail.com, James.Crosby@arm.com",
    description = ("Generate IDE project files to debug ELF files."),
    license = "Apache-2.0",
    keywords = "debug c cpp project generator embedded",
    url = "https://github.com/ARMmbed/valinor",
    packages=find_packages(),
    package_data={
    },
    long_description=read('docs/pypi.txt'),
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Topic :: Utilities",
        "License :: OSI Approved :: Apache Software License",
        "Environment :: Console",
    ],
    entry_points={
        "console_scripts": [
            "valinor=valinor:main",
        ],
    },
    test_suite = 'nose.collector',
    install_requires=[
        'pyyaml>=5.1,<6.0',
        'Jinja2>=2.7.0,<3',
        'setuptools',
        'colorama>=0.3,<0.4',
        'pyOCD>=0.3,<1.0',
        'project_generator>=0.8.0,<0.9.0',
        'pyelftools==0.23',
    ],
    tests_require=[
        'nose',
    ]
)



#!/usr/bin/env python
# Copyright 2014-2015 ARM Limited
#
# Licensed under the Apache License, Version 2.0
# See LICENSE file for details.


# standard library modules, , ,
import unittest
import os
import tempfile
import shutil

# internal modules:
from . import cli

class TestCLIOutputDirectory(unittest.TestCase):
    def setUp(self):
        self.workingdir = tempfile.mkdtemp()
        exedir = os.path.join(self.workingdir, 'some', 'long', 'path')
        os.makedirs(exedir)
        self.exe_path = os.path.join(exedir, 'myexe')
        with open(self.exe_path, 'w') as f:
            f.write('ELF')

    def tearDown(self):
        shutil.rmtree(self.workingdir)

    def runCheck(self, args):
        out, err, status = cli.run(args, cwd = self.workingdir)
        print(out)
        print(err)
        self.assertEqual(status, 0)
        return err or out

    def testUVision(self):
        def runWithDir(d=None):
            args = [
                '--tool', 'uvision',
                '--target', 'K64F',
                os.path.relpath(self.exe_path, self.workingdir),
                '-n'
            ]
            if d:
                args += ['-d', d]
            out = self.runCheck(args)

        runWithDir()
        self.assertTrue(os.path.isfile(self.exe_path + '.uvproj'))

        runWithDir('somesubdir')
        self.assertTrue(os.path.isfile(os.path.join(self.workingdir, 'somesubdir', 'myexe.uvproj')))

        runWithDir(os.path.join('somesubdir','anotherdir') + os.path.sep)
        self.assertTrue(os.path.isfile(os.path.join(self.workingdir, 'somesubdir', 'anotherdir', 'myexe.uvproj')))

    def testARMNoneEABIGDB(self):
        def runWithDir(d=None):
            args = [
                '--tool', 'arm_none_eabi_gdb',
                '--target', 'K64F',
                os.path.relpath(self.exe_path, self.workingdir),
                '-n'
            ]
            if d:
                args += ['-d', d]
            out = self.runCheck(args)

        runWithDir()
        self.assertTrue(os.path.isfile(self.exe_path + '.gdbstartup'))

        runWithDir('somesubdir')
        self.assertTrue(os.path.isfile(os.path.join(self.workingdir, 'somesubdir', 'myexe.gdbstartup')))

        runWithDir(os.path.join('somesubdir','anotherdir') + os.path.sep)
        self.assertTrue(os.path.isfile(os.path.join(self.workingdir, 'somesubdir', 'anotherdir', 'myexe.gdbstartup')))




